/* util.h -- definitions for gzip support

   Copyright (C) 2010 Gilles Buloz
   Copyright (C) 1997-1999, 2002, 2006, 2009 Free Software Foundation, Inc.
   Copyright (C) 1992-1993 Jean-loup Gailly

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef _UTIL_H_
#define _UTIL_H_

#include "FreeRTOS.h"
#include "queue.h"
#include "gzip.h"

typedef void *voidp;
typedef int file_t; /* Do not use stdio */

typedef unsigned char  uch;
typedef unsigned short ush;
typedef unsigned long  ulg;

#ifdef MESSAGES
#define msg_print(param...) fprintf(param)
#else
#define msg_print(param...)
#endif

#ifdef MAXSEG_64K
#  define fcalloc(items,size) calloc((items),(size))
#else
#  define fcalloc(items,size) malloc((size_t)(items)*(size_t)(size))
#endif
#define fcfree(ptr) free(ptr)

#define memzero(s, n) memset ((voidp)(s), 0, (n))

#define OS_CODE  0x03  /* assume Unix */
#define near
#define	GZIP_MAGIC "\037\213" /* Magic header for gzip files, 1F 8B */

/* Compression methods (see algorithm.doc) */
#define STORED      0
#define COMPRESSED  1
#define PACKED      2
#define LZHED       3
/* methods 4 to 7 reserved */
#define DEFLATED    8
extern int method;         /* compression method */

#define EXTERN(type, array)  extern type * near array
#define DECLARE(type, array, size)  type * near array
#ifdef DYN_ALLOC
#  define ALLOC(type, array, size, rc) {	       		      \
		array = (type*)fcalloc((size_t)(((size)+1L)/2), 2*sizeof(type)); \
		if (!array) {msg_print(stderr, "ERROR: allocating memory\n"); rc |= -1;} \
}
#  define FREE(array) {if (array != NULL) fcfree(array), array=NULL;}
#else
#  define ALLOC(type, array, size, rc) { \
		static type near _array[size]; \
		array = _array; \
}
#  define FREE(array)
#endif

extern unsigned insize;
extern unsigned inptr;
extern unsigned outcnt;

extern int test_mode;

extern xQueueHandle  ifd;
extern xQueueHandle  ofd;
extern int on_the_fly_state;
extern int on_the_fly_eofile;
extern char *on_the_fly_data_ptr;
extern off_t on_the_fly_data_size;

extern char *input_data_ptr;
extern char *output_data_ptr;
extern off_t input_data_size;
extern off_t output_max_size;

extern int input_error;
extern int output_error;

extern int level;        /* compression level (also used by trees.c) */

extern ulg crc;

/* Diagnostic functions */
#ifdef DEBUG
extern int verbose;
#  define Assert(cond,msg) {if (!(cond)) fprintf(stderr, "ERROR: %s\n", (msg));}
#  define Trace(x) fprintf x
#  define Tracev(x) {if (verbose) fprintf x ;}
#  define Tracevv(x) {if (verbose>1) fprintf x ;}
#  define Tracec(c,x) {if (verbose && (c)) fprintf x ;}
#  define Tracecv(c,x) {if (verbose>1 && (c)) fprintf x ;}
#else
#  define Assert(cond,msg)
#  define Trace(x)
#  define Tracev(x)
#  define Tracevv(x)
#  define Tracec(c,x)
#  define Tracecv(c,x)
#endif

/* in util.c: */
extern ulg updcrc OF((uch *s, unsigned n));

extern int readFromQueue OF((xQueueHandle fd, voidp buf, unsigned int cnt));
extern int writeEofToQueue OF((xQueueHandle fd));
extern int writeToQueue OF((xQueueHandle fd, voidp buf, unsigned int cnt));

extern void setLevel OF((int lvl));
extern int getLevel OF(());

#endif /* _UTIL_H_ */
