/*
 * wifly.h
 *
 *  Created on: 22/10/2012
 *      Author: Javier Rufas
 */

#ifndef WIFLY_H_
#define WIFLY_H_
#include "uart.h"


#define MAX_UART_IDLE_TIME 100UL //milliseconds from last received character

/*********************************************************************
 *
 * Constants and strings taken from:
 *
 * Arduino WiFly Device Driver
 * WiFlySerial v1.08
 * Copyright GPL 2.1 Tom Waldock 2011, 2012
 * http://arduinowifly.sourceforge.net/
 *
 *********************************************************************/



#define COMMAND_BUFFER_SIZE 64
#define SMALL_COMMAND_BUFFER_SIZE 20
#define RESPONSE_BUFFER_SIZE 80
#define INDICATOR_BUFFER_SIZE 16
#define IP_ADDR_WIDTH  18
#define UC_N_IP_BYTES  4      // number of bytes per IP Address (need test for IPv6)

// default ports as set in WiFly - if you change these, make certain you call setRemotePort/setLocalPort
#define WIFLY_DEFAULT_REMOTE_PORT 80
#define WIFLY_DEFAULT_LOCAL_PORT 2000
#define WIFLY_DEFAULT_BAUD_RATE 9600

#define COMMAND_MODE_GUARD_TIME 	300UL // in milliseconds.  Must be at least 250ms.
#define DEFAULT_WAIT_TIME			1500UL  // waiting time for a command
#define ATTN_WAIT_TIME				1000UL  // waiting time for a reponse after a $$$ attention signal
#define JOIN_WAIT_TIME 				10000UL  // joining a network could take longer
#define SCAN_WAIT_TIME 				20000UL  // scanning available networks could take longer
#define SERVING_WAIT_TIME			300000UL   // 5 minutes default wait time for serving a connection in milliseconds
#define WIFLY_RESTART_WAIT_TIME		2000UL  // wait time after issuing reboot command.
#define WIFLY_TIME_SYNC_INTERVAL 	100UL   // 15 minute interval between clock resyncs
#define WIFLY_NTP_SYNC_INTERVAL     "60"    // sync with NTP server every hour


#define COMMAND_MODE_ENTER_RETRY_ATTEMPTS 5
#define COMMAND_RETRY_ATTEMPTS 3

// WiFly Responses
#define PROMPT_NONE					0x0
#define PROMPT_EXPECTED_TOKEN_FOUND	0x1
#define PROMPT_READY 			   	0x2
#define PROMPT_CMD_MODE 	   		0x4
#define PROMPT_AOK 			    	0x8
#define PROMPT_OTHER 				0x10
#define PROMPT_CMD_ERR 				0x20
#define PROMPT_TIMEOUT 				0x40
#define PROMPT_OPEN 				0x80
#define PROMPT_CLOSE 				0x100
#define PROMPT_OPEN_ALREADY			0x200

#define N_PROMPTS 				9
#define WIFLY_MSG_EXPECTED  	0
#define WIFLY_MSG_AOK  			1
#define WIFLY_MSG_CMD  			2
#define WIFLY_MSG_ERR  			3
#define WIFLY_MSG_PROMPT  		4
#define WIFLY_MSG_PROMPT2  		5
#define WIFLY_MSG_CLOSE  		6
#define WIFLY_MSG_OPEN  		7
#define WIFLY_MSG_OPEN_ALREADY  8

// Auth Modes for Network Authentication
// See WiFly manual for details
#define WIFLY_AUTH_OPEN        0    // Open (default)
#define WIFLY_AUTH_WEP         1    // WEP-128
#define WIFLY_AUTH_WPA1        2    // WPA1
#define WIFLY_AUTH_WPA1_2      3    // Mixed-mode WPA1 and WPA2-PSK
#define WIFLY_AUTH_WPA2_PSK    4    // WPA2-PSK
#define WIFLY_AUTH_ADHOC       6    // Ad-hoc, join any Ad-hoc network

// Join modes for joining networks
// See WiFly manual for details
#define WIFLY_JOIN_MANUAL      0    // Manual, do not try to join automatically
#define WIFLY_JOIN_AUTO        1    // Try to join the access point that matches the stored SSID, passkey and channel. (Default)
#define WIFLY_JOIN_ANY         2    // Join ANY access point with security matching the stored authentication mode. Ignores stored SSID and searches for strongest signal.
#define WIFLY_JOIN_MAKE_ADHOC  4    // Create an Ad-Hoc network using stored SSID, IP and netmask.  Channel MUST be set. DHCP should be 0.

// DHCP modes
// See WiFly manual for details
#define WIFLY_DHCP_OFF         0    // DHCP OFF, use static IP address
#define WIFLY_DHCP_ON          1    // DHCP ON, get IP, gateway from Access Point.
#define WIFLY_DHCP_AUTO        2    // DHCP AUTO-IP, generally used with Adhoc networks
#define WIFLY_DHCP_CACHE       3    // DHCP CACHE, use previous IP address if lease is not expired.

// IP modes: UDP and TCP
#define WIFLY_IPMODE_UDP       0x02;    // Same as W5100_defn's SmMR::UDP
#define WIFLY_IPMODE_TCP       0x01;    // Same as W5100_defn's SmMR::TCP

#define WIFLY_STATUS_OFFLINE    0
#define WIFLY_STATUS_ONLINE     1





// Index of strings
#define STI_WIFLYDEVICE_INDEX_JOIN        0
#define STI_WIFLYDEVICE_INDEX_ASSOCIATED  1
#define STI_WIFLYDEVICE_ATTN              2
#define STI_WIFLYDEVICE_VER               3
#define STI_WIFLYDEVICE_GET_MAC           4
#define STI_WIFLYDEVICE_GET_IP            5
#define STI_WIFLYDEVICE_GET_GW            6
#define STI_WIFLYDEVICE_GET_NM            7
#define STI_WIFLYDEVICE_LEAVE             8
#define STI_WIFLYDEVICE_SET_SSID          9
#define STI_WIFLYDEVICE_SET_PASSPHRASE    10
#define STI_WIFLYDEVICE_NETWORK_SCAN      11
#define STI_WIFLYDEVICE_AOK               12
#define STI_WIFLYDEVICE_SET_UART_BAUD     13
#define STI_WIFLYDEVICE_DEAUTH            14
#define STI_WIFLYDEVICE_SET_NTP           15
#define STI_WIFLYDEVICE_SET_NTP_ENABLE    16
#define STI_WIFLYDEVICE_SET_DEVICEID      17
#define STI_WIFLYDEVICE_GET_IP_DETAILS    18
#define STI_WIFLYDEVICE_LEAVE_CMD_MODE    19
#define STI_WIFLYDEVICE_GET_DNS_DETAILS   20
#define STI_WIFLYDEVICE_GET_TIME          21
#define STI_WIFLYDEVICE_SET_DHCP          22
#define STI_WIFLYDEVICE_SET_IP            23
#define STI_WIFLYDEVICE_SET_NETMASK       24
#define STI_WIFLYDEVICE_SET_GATEWAY       25
#define STI_WIFLYDEVICE_SET_DNS           26
#define STI_WIFLYDEVICE_ERR_REBOOT        27
#define STI_WIFLYDEVICE_ERR_START_FAIL    28
#define STI_WIFLYDEVICE_SET_UART_MODE     29
#define STI_WIFLYDEVICE_GET_WLAN          30
#define STI_WIFLYDEVICE_GET_RSSI          31
#define STI_WIFLYDEVICE_GET_BATTERY       32
#define STI_WIFLYDEVICE_LIBRARY_VERSION   33
#define STI_WIFLYDEVICE_SET_CHANNEL       34
#define STI_WIFLYDEVICE_SET_WIFI_AUTH     35
#define STI_WIFLYDEVICE_SET_WIFI_JOIN     36
#define STI_WIFLYDEVICE_GET_STATUS        37
#define STI_WIFLYDEVICE_GET_MAC_ADDR      38
#define STI_WIFLYDEVICE_RETURN            39
#define STI_WIFLYDEVICE_GET_IP_IND        40
#define STI_WIFLYDEVICE_GET_NM_IND        41
#define STI_WIFLYDEVICE_GET_GW_IND        42
#define STI_WIFLYDEVICE_GET_DNS_IND       43
#define STI_WIFLYDEVICE_GET_WLAN_SSID_IND 44
#define STI_WIFLYDEVICE_GET_RSSI_IND      45
#define STI_WIFLYDEVICE_GET_BATTERY_IND   46
#define STI_WIFLYDEVICE_GET_WLAN_DEV_IND  47
#define STI_WIFLYDEVICE_GET_TIME_IND      48
#define STI_WIFLYDEVICE_GET_STATUS_IND    49
#define STI_WIFLYDEVICE_GET_IP_UP_IND     50
#define STI_WIFLYDEVICE_OPEN              51
#define STI_WIFLYDEVICE_REBOOT            52
#define STI_WIFLYDEVICE_CLOSE             53
#define STI_WIFLYDEVICE_SET_LOCAL_PORT    54
#define STI_WIFLYDEVICE_SET_REMOTE_PORT   55
#define STI_WIFLYDEVICE_SET_PROTOCOL      56
#define STI_WIFLYDEVICE_SAVE              57
#define STI_WIFLYDEVICE_GET_WLAN_TX_IND   58
#define STI_WIFLYDEVICE_SET_WLAN_TX       59


/*
 * Structure that holds the whole data of a WLAN
 */
struct _wlan {
	char *ssid;
	char *phrase;
	char *key;
	unsigned short mask;

	unsigned long auth:4,
	              channel:4,
	              rate:4,
	              tx:4,
	              join:3,
	              linkmon:3,
	              hide:1,
	              flag:1,
	              ext_antenna:1;
};

struct _wifly {
	struct _uart uart;
	//struct _wlan wlan;
	//char responseBuffer[COMMAND_BUFFER_SIZE];
	// internal buffer for command-prompt
	char    szWiFlyPrompt[INDICATOR_BUFFER_SIZE ];

	// Internal status flags
	long    fStatus;
	boolean bWiFlyInCommandMode;
	boolean bWiFlyConnectionOpen;
	boolean bDebugMode;
	char*   pControl;

	// Ports for connections
	int     iRemotePort;
	int     iLocalPort;
	long 	lUTC_Offset_seconds;

	unsigned long (*millis)();
	void (*Debug) (void *param, char *msg, int size, int mode);
	void *debugParam;
};


void WiFlyDelay(struct _wifly *wifly, int ms);
void WiFlyDelayUntilNoData(struct _wifly *wifly, int msMax);
void WiFlyDelayUntilReceivingData(struct _wifly *wifly, int msInactivity);
void WiFlyReceiveDataUntilDelay(struct _wifly *wifly, int msInactivity, byte *buffer, unsigned long bufSize);


boolean WiFlysetProtocol( struct _wifly  *wifly, unsigned int iProtocol);
boolean WiFlyopenConnection(struct _wifly  *wifly, const char* pURL, const unsigned long iWaitTime);
boolean WiFlycloseConnection(struct _wifly  *wifly, boolean bSafeClose);
boolean WiFlyserveConnection( struct _wifly  *wifly, const unsigned long reconnectWaitTime );
boolean WiFlyisTCPConnected(struct _wifly  *wifly);
boolean WiFlyisAssociated(struct _wifly  *wifly);
boolean WiFlyisAuthenticated(struct _wifly  *wifly);
boolean WiFlyisDNSfound(struct _wifly  *wifly);
boolean WiFlyisDNScontacted(struct _wifly  *wifly);
boolean WiFlyisInCommandMode(struct _wifly  *wifly);
boolean WiFlyisConnectionOpen(struct _wifly  *wifly);
boolean WiFlyisifUp(struct _wifly  *wifly);
boolean WiFlyleave(struct _wifly  *wifly);
boolean WiFlysetSSID( struct _wifly  *wifly, const char* pSSID);
boolean WiFlysetPassphrase( struct _wifly  *wifly, const char* pPassphrase);
boolean WiFlysetNTP(struct _wifly  *wifly, const char* pNTP);
boolean WiFlysetNTP_Update_Frequency(struct _wifly  *wifly, const char* pNTP_Update);
boolean WiFlysetNTP_UTC_Offset(struct _wifly  *wifly, float fltUTC_Offset_hours);
boolean WiFlysetDeviceID(struct _wifly *wifly,const char* pDeviceID);
boolean WiFlysetDHCPMode(struct _wifly  *wifly, const int iDHCPMode);
boolean WiFlysetIP( struct _wifly  *wifly, const char* pIP);
boolean WiFlysetNetMask( struct _wifly  *wifly, const char* pNM);
boolean WiFlysetGateway( struct _wifly  *wifly, const char* pGW);
boolean WiFlysetDNS( struct _wifly  *wifly, const char* pDNS);
boolean WiFlysetChannel( struct _wifly  *wifly, const char* pChannel);
boolean WiFlysetAuthMode( struct _wifly  *wifly, int iAuthMode);
boolean WiFlysetJoinMode( struct _wifly  *wifly, int iJoinMode);
boolean WiFlysetLocalPort( struct _wifly  *wifly, int iNewLocalPort);
boolean WiFlysetRemotePort( struct _wifly  *wifly, int iNewRemotePort);
boolean WiFlysetTxPower( struct _wifly  *wifly, int iNewPower);
boolean WiFlyjoin(struct _wifly  *wifly);
boolean WiFlyjoinSSID(struct _wifly  *wifly, char* pSSID);
char* WiFlygetMAC(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetIP(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetNetMask(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetGateway(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetDNS(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetSSID(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetRSSI(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetBattery(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetDeviceID(struct _wifly  *wifly, char* pbuf, int buflen);
char* WiFlygetLibraryVersion(struct _wifly  *wifly, char* pbuf, int buflen);
unsigned long WiFlygetTime(struct _wifly  *wifly);
long WiFlygetDeviceStatus(struct _wifly *wifly);
int WiFlygetChannel(struct _wifly  *wifly);
boolean WiFlybegin(struct _wifly *wifly);
char* WiflyshowNetworkScan( struct _wifly *wifly, char* pNetScan, const int buflen);
int WiFlydrain(struct _wifly  *wifly);
int WiFlygetTxPower(struct _wifly *wifly);
char * WiFlysendPing(struct _wifly *wifly, char* host, char* pbufResponse, int buflen);
boolean WiFlysendHTTPGet(struct _wifly *wifly, char* url, char* pbufResponse, int buflen, uint32_t waitTime);
unsigned long WiFlygetDataHTTPGet(struct _wifly *wifly, char* url, char* pbufResponse, int buflen, uint32_t waitTime);
boolean WiFlyexitCommandMode(struct _wifly *wifly);

char *WiFlygetFixedPrompt(const int PromptIndex);
char *WiFlygetDeviceString(const int StringIndex);
boolean WiFlysendCommandSimple(struct _wifly *wifly, char* pCommand, char* pSuccessIndicator);

#define DEBUG_MODE_FULL       30
#define DEBUG_MODE_INFO       20
#define DEBUG_MODE_CRITICAL   10
#define DEBUG_MODE_NONE       0

#endif /* WIFLY_H_ */
