/*
 * wifly.c
 *
 *  Created on: 22/10/2012
 *      Author: Javier Rufas
 */

#include "LPC17xx.h"
#include <stdlib.h>
#include <string.h>
#include "type.h"
#include "strutils.h"
#include "wifly.h"

/*
 * Define strings in program space
 * Change name
 */

char s_WIFLYDEVICE_LIBRARY_VERSION[] = "WiFlyLPC v0.01" ;
char s_WIFLYDEVICE_JOIN[] = "join " ;
char s_WIFLYDEVICE_OPEN[] = "open " ;
char s_WIFLYDEVICE_CLOSE[] = "close" ;
char s_WIFLYDEVICE_ASSOCIATED[] = "ssociated" ;
char s_WIFLYDEVICE_ATTN[] = "$$$";
char s_WIFLYDEVICE_VER[] = "ver" ;
char s_WIFLYDEVICE_LEAVE_CMD_MODE[] ="exit";
char s_WIFLYDEVICE_REBOOT[] ="reboot";
char s_WIFLYDEVICE_SAVE[] ="save";
char s_WIFLYDEVICE_GET_MAC[] =" get mac";
char s_WIFLYDEVICE_GET_MAC_ADDR[] ="Addr=";
char s_WIFLYDEVICE_GET_IP[] =" get ip";
char s_WIFLYDEVICE_GET_GW[] = " "; // "GW=";
char s_WIFLYDEVICE_GET_NM[] = " "; // "NM=";
char s_WIFLYDEVICE_LEAVE[] ="leave";
char s_WIFLYDEVICE_SET_SSID[] =" set wlan s ";
char s_WIFLYDEVICE_SET_CHANNEL[] =" set wlan c ";
char s_WIFLYDEVICE_SET_WIFI_AUTH[] =" set wlan a ";
char s_WIFLYDEVICE_SET_WIFI_JOIN[] =" set wlan j ";
char s_WIFLYDEVICE_SET_PASSPHRASE[] =" set w p ";
char s_WIFLYDEVICE_NETWORK_SCAN[] ="scan ";
char s_WIFLYDEVICE_AOK[] ="";
char s_WIFLYDEVICE_SET_UART_BAUD[] ="set u b 9600 ";
char s_WIFLYDEVICE_DEAUTH[] ="Deauth";
char s_WIFLYDEVICE_SET_NTP[] =" set time a ";
char s_WIFLYDEVICE_SET_NTP_ENABLE[] ="set time e ";
char s_WIFLYDEVICE_SET_DEVICEID[] ="set opt deviceid ";
char s_WIFLYDEVICE_IP_DETAILS[] ="get ip";
char s_WIFLYDEVICE_GET_DNS_DETAILS[] ="get dns";
char s_WIFLYDEVICE_GET_TIME[] ="show t t";
char s_WIFLYDEVICE_SET_DHCP[] ="set ip dhcp ";
char s_WIFLYDEVICE_SET_IP[] ="set ip a ";
char s_WIFLYDEVICE_SET_NETMASK[] ="set ip n ";
char s_WIFLYDEVICE_SET_GATEWAY[] ="set ip g ";
char s_WIFLYDEVICE_SET_DNS[] ="set dns addr ";
char s_WIFLYDEVICE_SET_LOCAL_PORT[] ="set ip local ";
char s_WIFLYDEVICE_SET_REMOTE_PORT[] ="set ip remote  ";
char s_WIFLYDEVICE_SET_PROTOCOL[] ="set ip proto ";
char s_WIFLYDEVICE_ERR_REBOOOT[] ="Attempting reboot";
char s_WIFLYDEVICE_ERR_START_FAIL[] ="Failed to get cmd prompt:Halted.";
char s_WIFLYDEVICE_SET_UART_MODE[] ="set u m 1 ";
char s_WIFLYDEVICE_GET_WLAN[] ="get wlan ";
char s_WIFLYDEVICE_GET_RSSI[] ="show rssi ";
char s_WIFLYDEVICE_GET_BATTERY[] ="show batt ";
char s_WIFLYDEVICE_GET_STATUS[] ="show conn ";
char s_WIFLYDEVICE_RETURN[] ="\r";
char s_WIFLYDEVICE_GET_IP_IND[] ="IP=";
char s_WIFLYDEVICE_GET_NM_IND[] ="NM=";
char s_WIFLYDEVICE_GET_GW_IND[] ="GW=";
char s_WIFLYDEVICE_GET_DNS_IND[] ="DNS=";
char s_WIFLYDEVICE_GET_WLAN_SSID_IND[] ="SSID=";
char s_WIFLYDEVICE_GET_RSSI_IND[] ="RSSI=";
char s_WIFLYDEVICE_GET_WLAN_DEV_IND[] ="DeviceID=";
char s_WIFLYDEVICE_GET_BATTERY_IND[] ="Batt=";
char s_WIFLYDEVICE_GET_TIME_IND[] ="RTC=";
char s_WIFLYDEVICE_GET_STATUS_IND[] ="8";
char s_WIFLYDEVICE_GET_IP_UP_IND[] ="F=";
char s_WIFLYDEVICE_GET_WLAN_TX_IND[] ="TxPower=";
char s_WIFLYDEVICE_SET_WLAN_TX[] ="set wlan t ";

// String Table in Program space
const char *WiFlyDevice_string_table[] =
{
  // 0-based index, see STI_WIFLY_DEVICE_ list above.
  s_WIFLYDEVICE_JOIN,
  s_WIFLYDEVICE_ASSOCIATED,
  s_WIFLYDEVICE_ATTN,
  s_WIFLYDEVICE_VER,
  s_WIFLYDEVICE_GET_MAC,
  s_WIFLYDEVICE_GET_IP,
  s_WIFLYDEVICE_GET_GW,
  s_WIFLYDEVICE_GET_NM,
  s_WIFLYDEVICE_LEAVE,
  s_WIFLYDEVICE_SET_SSID,
  // 10 follows
  s_WIFLYDEVICE_SET_PASSPHRASE,
  s_WIFLYDEVICE_NETWORK_SCAN,
  s_WIFLYDEVICE_AOK,
  s_WIFLYDEVICE_SET_UART_BAUD,
  s_WIFLYDEVICE_DEAUTH,
  s_WIFLYDEVICE_SET_NTP,
  s_WIFLYDEVICE_SET_NTP_ENABLE,
  s_WIFLYDEVICE_SET_DEVICEID,
  s_WIFLYDEVICE_IP_DETAILS,
  s_WIFLYDEVICE_LEAVE_CMD_MODE,
  // 20 follows
  s_WIFLYDEVICE_GET_DNS_DETAILS,
  s_WIFLYDEVICE_GET_TIME,
  s_WIFLYDEVICE_SET_DHCP,
  s_WIFLYDEVICE_SET_IP,
  s_WIFLYDEVICE_SET_NETMASK,
  s_WIFLYDEVICE_SET_GATEWAY,
  s_WIFLYDEVICE_SET_DNS,
  s_WIFLYDEVICE_ERR_REBOOOT,
  s_WIFLYDEVICE_ERR_START_FAIL,
  s_WIFLYDEVICE_SET_UART_MODE,
  // 30 follows
  s_WIFLYDEVICE_GET_WLAN,
  s_WIFLYDEVICE_GET_RSSI,
  s_WIFLYDEVICE_GET_BATTERY,
  s_WIFLYDEVICE_LIBRARY_VERSION,
  s_WIFLYDEVICE_SET_CHANNEL,
  s_WIFLYDEVICE_SET_WIFI_AUTH,
  s_WIFLYDEVICE_SET_WIFI_JOIN,
  s_WIFLYDEVICE_GET_STATUS,
  s_WIFLYDEVICE_GET_MAC_ADDR,
  s_WIFLYDEVICE_RETURN,
  // 40 follows
  s_WIFLYDEVICE_GET_IP_IND,
  s_WIFLYDEVICE_GET_NM_IND,
  s_WIFLYDEVICE_GET_GW_IND,
  s_WIFLYDEVICE_GET_DNS_IND,
  s_WIFLYDEVICE_GET_WLAN_SSID_IND,
  s_WIFLYDEVICE_GET_RSSI_IND,
  s_WIFLYDEVICE_GET_BATTERY_IND,
  s_WIFLYDEVICE_GET_WLAN_DEV_IND,
  s_WIFLYDEVICE_GET_TIME_IND,
  s_WIFLYDEVICE_GET_STATUS_IND,
  // 50 follows
  s_WIFLYDEVICE_GET_IP_UP_IND,
  s_WIFLYDEVICE_OPEN,
  s_WIFLYDEVICE_REBOOT,
  s_WIFLYDEVICE_CLOSE,
  s_WIFLYDEVICE_SET_LOCAL_PORT,
  s_WIFLYDEVICE_SET_REMOTE_PORT,
  s_WIFLYDEVICE_SET_PROTOCOL,
  s_WIFLYDEVICE_SAVE,
  s_WIFLYDEVICE_GET_WLAN_TX_IND,
  s_WIFLYDEVICE_SET_WLAN_TX
  // 60 follows
};

// WiFly-specific prompt codes
static char* WiFlyFixedPrompts[N_PROMPTS] = { "","AOK", "CMD", "ERR: ?", "",">","*CLOS*","*OPEN*","ERR:Connected" };
static int WiFlyFixedFlags[N_PROMPTS] = {PROMPT_EXPECTED_TOKEN_FOUND, PROMPT_AOK, PROMPT_CMD_MODE, PROMPT_CMD_ERR, PROMPT_READY, PROMPT_READY,PROMPT_CLOSE, PROMPT_OPEN, PROMPT_OPEN_ALREADY};


// Internal functions declaration
int ScanForPattern(struct _wifly *wifly, char* responseBuffer, const int buflen, const char *pExpectedPrompt,
		const boolean bCollecting, const unsigned long WaitTime,
		const boolean bPromptAfterResult);
boolean StartCommandMode(struct _wifly *wifly, char* pBuffer, const int bufSize);
boolean SendCommand(struct _wifly *wifly, char *pCmd, char *SuccessIndicator,
		char* pResultBuffer, const int bufsize, const boolean bCollecting,
		const unsigned long iWaitTime, const boolean bClear,
		const boolean bPromptAfterResult);
boolean SendCommand2(struct _wifly *wifly, char *pCmd, char *pParam, char *SuccessIndicator,
		char* pResultBuffer, const int bufsize, const boolean bCollecting,
		const unsigned long iWaitTime, const boolean bClear,
		const boolean bPromptAfterResult);
boolean GetCmdPrompt(struct _wifly *wifly);
//void WiFlygetDeviceStatus(struct _wifly *wifly);
char* ExtractDetailIdx(struct _wifly *wifly, const int idxCommand,
		char* pDetail, int buflen, const int idxSearch, const int idxStop);
char* ExtractDetail(struct _wifly *wifly, char* pCommand, char* pDetail,
		int buflen, const char* pFrom, const char* pTo);
void reboot(struct _wifly *wifly);
boolean SendInquiry(struct _wifly *wifly, char *Command, char* pBuffer,
		const int bufsize);
boolean SendInquirySimple(struct _wifly *wifly, char *Command);

char* ExtractLineFromBuffer(struct _wifly  *wifly, const int idString,  char* pBuffer, const int bufsize, const char* pStartPattern, const char* pTerminator);
boolean issueSetting( struct _wifly  *wifly, int idxCommand, const char* pParam);

void DebugPrint( struct _wifly  *wifly, char* pParam, int mode);
void DebugPrintReceive( struct _wifly  *wifly, char* pParam, int mode);
void DebugPrintSend( struct _wifly  *wifly, char* pParam, int mode);
void DebugPrintChar( struct _wifly  *wifly, char* pParam, int len, int mode);


// Delay in milliseconds
// depends of millis() function in wifly structure.
void WiFlyDelay(struct _wifly *wifly, int ms)
{
	uint32_t time = wifly->millis();
	while((wifly->millis()- time) < ms);
}

// Delay until no data in UART
// depends of millis() function.
void WiFlyDelayUntilNoData(struct _wifly *wifly, int msMax)
{
	uint32_t time = wifly->millis();
	struct _uart* uart = &(wifly->uart);

	while((wifly->millis()- time) < msMax ){
		 if (UARTisAvailable(uart)) return;
	}
}

// Delay until UART receiving data
// depends of millis() function.
void WiFlyDelayUntilReceivingData(struct _wifly *wifly, int msInactivity)
{
	struct _uart* uart = &(wifly->uart);
	uint32_t count;

	// wait for a ms of inactivity
	if(UARTisAvailable(uart)){
		do {
			count = *(uart->pUARTCount);
			WiFlyDelay(wifly, msInactivity);
		} while (count > *(uart->pUARTCount)); //while count increments
	}
}

// Receive Data considering a Delay of inactivity
// depends of millis() function.
void WiFlyReceiveDataUntilDelay(struct _wifly *wifly, int msInactivity, byte *buffer, unsigned long bufSize)
{
	struct _uart* uart = &(wifly->uart);
	boolean pass =  false;

	do {
		if(UARTisAvailable(uart)) {
			UARTreadByte(uart, buffer);
			buffer++;
			pass=false;
		} else {
			if(pass) break;
			WiFlyDelay(wifly, msInactivity);
			pass = true;
		}
	}while(1);
}

// GetBuffer_P
// Returns pointer to a supplied Buffer
// If NULL, returns the read-only buffer

char* GetBuffer_P(const int StringIndex, char* pBuffer, int bufSize) {

	char *string = (char*) WiFlyDevice_string_table[StringIndex];

	if (pBuffer != NULL) {
		  memset(pBuffer, '\0', bufSize);
		  strncpy(pBuffer, string, bufSize);
		  string = pBuffer;
	}

	return string;
}

#define MS_TIME    SystemCoreClock/1000    /* 1 ms */

unsigned long defaultMillis()
{
	return 0UL;
}

// ScanForPattern
//
// General-purpose stream watcher.
// Monitors incoming stream until given prompt is detected, or error conditions, or until timer expired
////
// Parameters
// ResponseBuffer    buffer for WiFly response
// bufsize           size of buffer
// pExpectedPrompt   Marker to find
// bCollecting       true: collect chars in buffer UNTIL marker found, false: discard UNTIL marker found
// WaitTime          Timeout duration to wait for response
// bPromptAfterResult true: version prompt after result, false: version prompt precedes results (scan, join).
//
// Returns: (see .h file)  OR-ed flags of the following
// WiFly Responses:
//#define PROMPT_NONE 0
//#define PROMPT_EXPECTED_TOKEN_FOUND 1
//#define PROMPT_READY 2
//#define PROMPT_CMD_MODE 4
//#define PROMPT_AOK 8
//#define PROMPT_OTHER 16
//#define PROMPT_CMD_ERR 32
//#define PROMPT_TIMEOUT 64
//#define PROMPT_OPEN 128
//#define PROMPT_CLOSE 256

int ScanForPattern(struct _wifly *wifly, char* responseBuffer,
		const int buflen, const char *pExpectedPrompt,
		const boolean bCollecting, const unsigned long WaitTime,
		const boolean bPromptAfterResult) {

	int iPromptFound = PROMPT_NONE;
	char chResponse = 'A';
	int bufpos = 0;
	int bufsize = buflen - 1; //terminating null for bufsize
	boolean bWaiting = true;
	char* pFixedCurrent[N_PROMPTS];
	int iFixedPrompt = 0;
	int i;

	WiFlyFixedPrompts[WIFLY_MSG_EXPECTED] = (char*) pExpectedPrompt;
	WiFlyFixedPrompts[WIFLY_MSG_PROMPT] = (char*) wifly->szWiFlyPrompt;

	DebugPrint(wifly, "ScanForPattern (", DEBUG_MODE_INFO);
	DebugPrint(wifly, (char*) pExpectedPrompt, DEBUG_MODE_INFO);
	DebugPrint(wifly, ")\n", DEBUG_MODE_INFO);

	for (i = 0; i < N_PROMPTS; i++) {
		pFixedCurrent[i] = WiFlyFixedPrompts[i];
	}

	memset(responseBuffer, '\0', buflen);
	unsigned long TimeAtStart = wifly->millis(); // capture current time

	while (bWaiting) {
		if (UARTisAvailable(&(wifly->uart))) {
			UARTreadByte(&(wifly->uart), (byte*) &chResponse);

			DebugPrintChar(wifly, &chResponse, 1, DEBUG_MODE_FULL);

			if (bCollecting) {
				responseBuffer[bufpos] = chResponse;
				if (++bufpos == bufsize) {
					bufpos = 0;
				} // if buffer wrapped
			} // if capturing

			for (iFixedPrompt = 0; iFixedPrompt < N_PROMPTS; iFixedPrompt++) {
				if (chResponse == *pFixedCurrent[iFixedPrompt]) {
					// deal with 'open' and 'scan' version-prompt appearing BEFORE result; ignore it
					if ((!bPromptAfterResult) && (iFixedPrompt
							== WIFLY_MSG_PROMPT || iFixedPrompt
							== WIFLY_MSG_PROMPT2) /* standard version-prompt */) {
						bWaiting = true;
						iPromptFound |= PROMPT_READY;
					} else {
						bWaiting
						= (*(++pFixedCurrent[iFixedPrompt]) == '\0' ? false
								: true); // done when end-of-string encountered.
						if (!bWaiting) {
							iPromptFound |= WiFlyFixedFlags[iFixedPrompt]; // if a prompt found then grab its flag.
						}
					} // handle prompt-BEFORE-result case
				} else {
					pFixedCurrent[iFixedPrompt]
					              = WiFlyFixedPrompts[iFixedPrompt]; // not next char expected; reset to beginning of string.
				} // if tracking expected response

/**************************************************************************
 * Prueba para resetear antes de comprobar. No entiendo que no funcione.
 **************************************************************************
				if (chResponse != *pFixedCurrent[iFixedPrompt]) { // not next char expected; reset to beginning of string.
					pFixedCurrent[iFixedPrompt] = WiFlyFixedPrompts[iFixedPrompt];
				}
				if (chResponse == *pFixedCurrent[iFixedPrompt]) {
					// deal with 'open' and 'scan' version-prompt appearing BEFORE result; ignore it
					if ((!bPromptAfterResult) && (iFixedPrompt
							== WIFLY_MSG_PROMPT || iFixedPrompt
							== WIFLY_MSG_PROMPT2) ) { // standard version-prompt
						bWaiting = true;
						iPromptFound |= PROMPT_READY;
					} else {
						bWaiting = (*(++pFixedCurrent[iFixedPrompt]) == '\0' ? false : true); // done when end-of-string encountered.
						if (!bWaiting) {
							iPromptFound |= WiFlyFixedFlags[iFixedPrompt]; // if a prompt found then grab its flag.
						}
					} // handle prompt-BEFORE-result case
				} // if tracking expected response
 ***************************************************************************/
		}
			// If the *OPEN* signal caught then a connection was opened.
			if (iPromptFound & (PROMPT_OPEN | PROMPT_OPEN_ALREADY)) {
				wifly->bWiFlyConnectionOpen = true;
				wifly->bWiFlyInCommandMode = false;
				iPromptFound &= (~WiFlyFixedFlags[WIFLY_MSG_CLOSE]); // clear prior close
			}

			// If the *CLOS* signal caught then a connection was closed
			// and we dropped into command mode
			if (iPromptFound & PROMPT_CLOSE) {
				wifly->bWiFlyConnectionOpen = false;
				wifly->bWiFlyInCommandMode = true;
				iPromptFound &= (~WiFlyFixedFlags[WIFLY_MSG_OPEN]); // clear prior open
			}

		} // if anything in uart
		// did we time-out?
		if ((wifly->millis() - TimeAtStart) >= WaitTime) {
			bWaiting = false;
		}
	} // while waiting for a line
	// could capture and compare with known prompt
	if (bCollecting) {
		responseBuffer[bufpos] = '\0';
		DebugPrint(wifly, "responseBuffer:", DEBUG_MODE_INFO);
		DebugPrint(wifly, responseBuffer, DEBUG_MODE_INFO);
		DebugPrint(wifly, "\n", DEBUG_MODE_INFO);
	}
	return (int) iPromptFound;

} // ScanForPattern


// The same that ScanForPattern, but searching pExpectedPrompt only
int ScanForSinglePattern(struct _wifly *wifly, char* responseBuffer,
		const int buflen, const char *pExpectedPrompt,
		const boolean bCollecting, const unsigned long WaitTime,
		const boolean bPromptAfterResult) {

	byte iPromptFound = PROMPT_NONE;
	char chResponse = 'A';
	int bufpos = 0;
	int bufsize = buflen - 1; //terminating null for bufsize
	boolean bWaiting = true;
	char* pFixedCurrent[N_PROMPTS];
	int iFixedPrompt = 0;

	WiFlyFixedPrompts[WIFLY_MSG_EXPECTED] = (char*) pExpectedPrompt;
	WiFlyFixedPrompts[WIFLY_MSG_PROMPT] = (char*) wifly->szWiFlyPrompt;

	DebugPrint(wifly, "ScanForSinglePattern (", DEBUG_MODE_INFO);
	DebugPrint(wifly, (char*) pExpectedPrompt, DEBUG_MODE_INFO);
	DebugPrint(wifly, ")\n", DEBUG_MODE_INFO);

	pFixedCurrent[0] = WiFlyFixedPrompts[0];

	memset(responseBuffer, '\0', bufsize);
	unsigned long TimeAtStart = wifly->millis(); // capture current time

	while (bWaiting) {
		if (UARTisAvailable(&(wifly->uart))) {
			UARTreadByte(&(wifly->uart), (byte*) &chResponse);

			DebugPrintChar(wifly, &chResponse, 1, DEBUG_MODE_FULL);

			if (bCollecting) {
				responseBuffer[bufpos] = chResponse;
				if (++bufpos == bufsize) {
					bufpos = 0;
				} // if buffer wrapped
			} // if capturing

			for (iFixedPrompt = 0; iFixedPrompt < 1; iFixedPrompt++) {
				if (chResponse == *pFixedCurrent[iFixedPrompt]) {
					// deal with 'open' and 'scan' version-prompt appearing BEFORE result; ignore it
					if ((!bPromptAfterResult) && (iFixedPrompt
							== WIFLY_MSG_PROMPT || iFixedPrompt
							== WIFLY_MSG_PROMPT2) /* standard version-prompt */) {
						bWaiting = true;
						iPromptFound |= PROMPT_READY;
					} else {
						bWaiting
						= (*(++pFixedCurrent[iFixedPrompt]) == '\0' ? false
								: true); // done when end-of-string encountered.
						if (!bWaiting) {
							iPromptFound |= WiFlyFixedFlags[iFixedPrompt]; // if a prompt found then grab its flag.
						}
					} // handle prompt-BEFORE-result case
				} else {
					pFixedCurrent[iFixedPrompt]
					              = WiFlyFixedPrompts[iFixedPrompt]; // not next char expected; reset to beginning of string.
				} // if tracking expected response
			}

		} // if anything in uart
		// did we time-out?
		if ((wifly->millis() - TimeAtStart) >= WaitTime) {
			bWaiting = false;
		}
	} // while waiting for a line
	// could capture and compare with known prompt
	if (bCollecting) {
		responseBuffer[bufpos] = '\0';
		DebugPrint(wifly, "responseBuffer:", DEBUG_MODE_INFO);
		DebugPrint(wifly, responseBuffer, DEBUG_MODE_INFO);
		DebugPrint(wifly, "\n", DEBUG_MODE_INFO);
	}
	return (int) iPromptFound;

} // ScanForSinglePattern


// Start Command Mode
//
// Attempt up to 5 times
// test is "Get a command prompt matching results of 'ver' command".
// if InCommand mode, try a 'cr'.
// If no useful result, assume not actually in command mode, force with $$$
//
// Returns true for Command mode entered, false if not (something weird afoot).

boolean StartCommandMode(struct _wifly *wifly, char* pBuffer, const int bufSize) {
	byte iPromptResult = 0;
	char* responseBuffer;
	boolean bWaiting = true;
	int nTries = 0;

	DebugPrint(wifly, "StartCommandMode()\n", DEBUG_MODE_INFO);

	if (pBuffer == NULL) {
		responseBuffer = (char*) malloc(bufSize); // defaults to COMMAND_BUFFER_SIZE
	} else {
		responseBuffer = pBuffer;
	}

	unsigned long TimeOutTime = wifly->millis() + ATTN_WAIT_TIME;
	// check if actually in command mode:
	while (!wifly->bWiFlyInCommandMode || bWaiting) {

		// if not effectively in command mode, try $$$
		if (!wifly->bWiFlyInCommandMode) {

			// Send $$$ , wait a moment, look for CMD
			WiFlyDelay(wifly, COMMAND_MODE_GUARD_TIME);

			DebugPrintSend(wifly, (char*) WiFlyDevice_string_table[STI_WIFLYDEVICE_ATTN], DEBUG_MODE_INFO);

			UARTsendString(&(wifly->uart),
					(byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_ATTN]);
			UARTflush(&(wifly->uart));
			WiFlyDelay(wifly, COMMAND_MODE_GUARD_TIME);
			if (nTries >= 2) {
				UARTsendString(&(wifly->uart), (byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_RETURN]);
				UARTflush(&(wifly->uart));
			}
			// expect CMD without a cr
			// WiFlyFixedPrompts[WIFLY_MSG_CMD]
			iPromptResult = ScanForPattern(wifly, responseBuffer, bufSize,
					WiFlyFixedPrompts[WIFLY_MSG_CMD], true, ATTN_WAIT_TIME, true);

			if (iPromptResult & (PROMPT_EXPECTED_TOKEN_FOUND | PROMPT_READY
					| PROMPT_CMD_MODE | PROMPT_CMD_ERR)) {
				wifly->bWiFlyInCommandMode = true;
				bWaiting = false;
			} else {
				wifly->bWiFlyInCommandMode = false;
			} // if one of several indicators of command-mode received.

		} else {

			// think we are in a command-mode - try a cr, then add a version command to get through.
			// send a ver + cr, should see a prompt.
			if (nTries > 2) {
				UARTsendString(
						&(wifly->uart),
						(byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_VER]);
				UARTflush(&(wifly->uart));
			}

			DebugPrint(wifly, "***scm:InCommandMode***\n", DEBUG_MODE_INFO);
			UARTsendString(&(wifly->uart), (byte*) "\r\r");
			UARTflush(&(wifly->uart));
			// bring in a cr-terminated line
			// wait for up to time limit for a cr to flow by
			iPromptResult = ScanForPattern(wifly, responseBuffer, bufSize,
					wifly->szWiFlyPrompt, false, DEFAULT_WAIT_TIME, true);
			// could have timed out, or have *READY*, CMD or have a nice CR.

			if (iPromptResult & (PROMPT_EXPECTED_TOKEN_FOUND | PROMPT_AOK
					| PROMPT_READY | PROMPT_CMD_MODE | PROMPT_CMD_ERR)) {
				wifly->bWiFlyInCommandMode = true;
				bWaiting = false;
			} else {
				wifly->bWiFlyInCommandMode = false;
			} // if one of several indicators of command-mode received.

		} //  else in in command command mode

		if (wifly->millis() >= TimeOutTime) {
			bWaiting = false;
		}
		nTries++;
	} // while trying to get into command mode

	// clean up as needed
	if (pBuffer == NULL) {
		free(responseBuffer);
	}
	return wifly->bWiFlyInCommandMode;
}

// exitCommandMode
// Exits from WiFly command mode.
//
// Watch the NSS for further traffic.
//
// Parameters:
// None
// Returns true on command success, false on failure.
boolean WiFlyexitCommandMode(struct _wifly *wifly) {

	char szCmd[INDICATOR_BUFFER_SIZE]; // exit command is short
	char szPrompt[INDICATOR_BUFFER_SIZE]; // exit Prompt is short:  EXIT (which looks like 'exit' but in upper case).
	char szResponse[INDICATOR_BUFFER_SIZE]; // small buffer for result

	strncpy(szCmd, WiFlyDevice_string_table[STI_WIFLYDEVICE_LEAVE_CMD_MODE],
			INDICATOR_BUFFER_SIZE);
	strncpy(szPrompt, WiFlyDevice_string_table[STI_WIFLYDEVICE_LEAVE_CMD_MODE],
			INDICATOR_BUFFER_SIZE);

	DebugPrint(wifly, "exitCommandMode\n", DEBUG_MODE_INFO);
	DebugPrintSend(wifly, szCmd, DEBUG_MODE_INFO);

	wifly->bWiFlyInCommandMode = !SendCommand(wifly, szCmd, strupr(szPrompt),
			szResponse, INDICATOR_BUFFER_SIZE, false, DEFAULT_WAIT_TIME, true,
			true);

	wifly->bWiFlyInCommandMode = false;
	return wifly->bWiFlyInCommandMode;
}

// SendCommand
// Issues a command to the WiFly device
// Captures results in Returned result
//
//
// Parameters:
// Command          The inquiry-command to send
// SuccessIndicator String to indicate success
// pResultBuffer    A place to put results of the command
// bufsize          Length of the pResultBuffer
// bCollecting      true = collect results, false=ignore results.
// iWaitTime        Time in milliseconds to wait for a result.
// bClear           true = drain any preceeding and subsequent characters, false=ignore
// bPromptAfterResult true=commands end with a version-prompt, false=version-prompt precedes results.
//
// Returns true on SuccessIndicator presence, false if absent.
boolean SendCommand2(struct _wifly *wifly, char *pCmd, char *pParam, char *SuccessIndicator,
		char* pResultBuffer, const int bufsize, const boolean bCollecting,
		const unsigned long iWaitTime, const boolean bClear,
		const boolean bPromptAfterResult) {

	boolean bCommandOK = false;
	char ch;
	int iResponse = 0;
	int iTry = 0;

	char* Command = pCmd;
	if (pCmd == pResultBuffer) {
		Command = (char*) malloc(sizeof(pCmd) + 1);
		strcpy(Command, pCmd);
	}

	DebugPrint(wifly, "SendCommand(", DEBUG_MODE_INFO);
	DebugPrint(wifly, pCmd, DEBUG_MODE_INFO);
	DebugPrint(wifly, ")\n", DEBUG_MODE_INFO);

	//
	// clear out leftover characters coming in

	if (bClear) {
		// DebugPrint("Clearing:", DEBUG_MODE_INFO);
		while (UARTisAvailable(&(wifly->uart))) {
			UARTreadByte(&(wifly->uart), (byte *) &ch);
			DebugPrintChar(wifly, &ch, 1, DEBUG_MODE_FULL);
			// DebugPrint( ch, DEBUG_MODE_FULL);
		}
	}

	if (StartCommandMode(wifly, pResultBuffer, bufsize)) {

		while (((iResponse & PROMPT_EXPECTED_TOKEN_FOUND)
				!= PROMPT_EXPECTED_TOKEN_FOUND) && iTry
				< COMMAND_RETRY_ATTEMPTS) {
			UARTflush(&(wifly->uart));
			DebugPrintSend(wifly, Command, DEBUG_MODE_INFO);
			UARTsendString(&(wifly->uart), (byte*) Command);
			if(pParam != NULL) {
				DebugPrintSend(wifly, pParam, DEBUG_MODE_INFO);
				UARTsendString(&(wifly->uart), (byte*) pParam);
			}
			UARTsendString(&(wifly->uart), (byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_RETURN]);
			UARTflush(&(wifly->uart));
			iResponse = ScanForPattern(wifly, pResultBuffer, bufsize,
					SuccessIndicator, bCollecting, iWaitTime,
					bPromptAfterResult);
			iTry++;
		}

	}
	if (pCmd == pResultBuffer) {
		free(Command);
	}

	if (bClear) {
		ScanForPattern(wifly, strchr(pResultBuffer, '\0') + 1,
				bufsize - strlen(pResultBuffer) - 1,
				WiFlyFixedPrompts[WIFLY_MSG_CLOSE], false, DEFAULT_WAIT_TIME,
				true);
		// DebugPrint("Clearing:", DEBUG_MODE_INFO);
		while (UARTisAvailable(&(wifly->uart))) {
			UARTreadByte(&(wifly->uart), (byte *) &ch);
			DebugPrintChar(wifly, &ch, 1, DEBUG_MODE_FULL);
			// DebugPrint( ch, DEBUG_MODE_FULL);
		}
	} // clear out leftover characters

	bCommandOK = (((iResponse & PROMPT_EXPECTED_TOKEN_FOUND)
			== PROMPT_EXPECTED_TOKEN_FOUND) ? true : false);

	return bCommandOK;
}

boolean SendCommand(struct _wifly *wifly, char *pCmd, char *SuccessIndicator,
		char* pResultBuffer, const int bufsize, const boolean bCollecting,
		const unsigned long iWaitTime, const boolean bClear,
		const boolean bPromptAfterResult) {

	return SendCommand2(wifly, pCmd, NULL, SuccessIndicator, pResultBuffer, bufsize, bCollecting,iWaitTime,  bClear, bPromptAfterResult);
}



// convenient and version with own small ignored response buffer.
boolean WiFlysendCommandSimple(struct _wifly *wifly, char* pCommand,
		char* pSuccessIndicator) {

	char bufResponse[INDICATOR_BUFFER_SIZE];

	return SendCommand(wifly, pCommand, pSuccessIndicator, bufResponse,
			INDICATOR_BUFFER_SIZE, false, DEFAULT_WAIT_TIME, false, true);
}

// GetCmdPrompt
// Obtains the WiFly command prompt string for use by other command functions.
// Parameters: None
// Sets global szWiFlyPrompt
// Returns command prompt on success or empty string on failure
boolean GetCmdPrompt(struct _wifly *wifly) {

	boolean bOk = false;
	char responseBuffer[RESPONSE_BUFFER_SIZE];

	DebugPrint(wifly, "GetCmdPrompt()\n", DEBUG_MODE_INFO);

	if (StartCommandMode(wifly, responseBuffer, RESPONSE_BUFFER_SIZE)) {
		UARTsendString(&(wifly->uart),
				(byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_VER]);
		UARTflush(&(wifly->uart));
		UARTsendString(&(wifly->uart), (byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_RETURN]);
		UARTflush(&(wifly->uart));
		ScanForPattern(wifly, responseBuffer, RESPONSE_BUFFER_SIZE,
				WiFlyFixedPrompts[WIFLY_MSG_PROMPT2], true,
				COMMAND_MODE_GUARD_TIME, true);
		char* pPromptStart = strrchr(responseBuffer, '<');
		char* pPromptEnd = strrchr(responseBuffer, '>');

		if ((pPromptStart < pPromptEnd) && pPromptStart && pPromptEnd) {
			strncpy(wifly->szWiFlyPrompt, pPromptStart,
					(size_t) (pPromptEnd - pPromptStart) + 1);
			wifly->szWiFlyPrompt[(pPromptEnd - pPromptStart) + 1] = '\0';
		}
	}

	if (strlen(wifly->szWiFlyPrompt) > 1) {
		bOk = true;
	} else {
		bOk = false;
	}
	return bOk;
}

// begin
// Initializes WiFly interface and starts communication with WiFly device.
//
// Parameters: none.
// Returns: true on initialize success, false on failure.
boolean WiFlybegin(struct _wifly *wifly) {
	boolean bStart = false;
	char szCmd[SMALL_COMMAND_BUFFER_SIZE];
	char szResponse[COMMAND_BUFFER_SIZE];
	//  char szIndicator[INDICATOR_BUFFER_SIZE];

	DebugPrint(wifly, "WiFlybegin()\n", DEBUG_MODE_INFO);

	if( wifly->millis == NULL) wifly->millis = defaultMillis;
	wifly->uart.baudrate = WIFLY_DEFAULT_BAUD_RATE;
	UARTinit(&(wifly->uart));
	UARTflush(&(wifly->uart));

	wifly->bWiFlyInCommandMode = false;
	wifly->bWiFlyConnectionOpen = false;
	wifly->fStatus = WIFLY_STATUS_OFFLINE ;
	strcpy(wifly->szWiFlyPrompt, WiFlyFixedPrompts[WIFLY_MSG_PROMPT2] );  // ">"

	wifly->iLocalPort = WIFLY_DEFAULT_LOCAL_PORT;
	wifly->iRemotePort = WIFLY_DEFAULT_REMOTE_PORT;

	  // default is UTC timezone
	wifly->lUTC_Offset_seconds = 0;

	wifly->pControl = WiFlyFixedPrompts[WIFLY_MSG_CLOSE];


	//Device may or may not be:
	// awake / asleep
	// net-connected / connection lost
	// IP assigned / no IP
	// in command mode / data mode
	// in known state / confused
	// Start by setting command prompt.
	//InitSysTickTimer(1000); // Initialize system tick timer to milliseconds


	wifly->bWiFlyInCommandMode = false;
	StartCommandMode(wifly, szCmd, SMALL_COMMAND_BUFFER_SIZE);

	// turn off echo
	// set baud rate

	bStart = SendCommand(wifly,
			(char *) WiFlyDevice_string_table[STI_WIFLYDEVICE_SET_UART_MODE],
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szResponse, COMMAND_BUFFER_SIZE,
			true, DEFAULT_WAIT_TIME, true, true);

	bStart = SendCommand(wifly,
			(char *) WiFlyDevice_string_table[STI_WIFLYDEVICE_SET_UART_BAUD],
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szResponse, COMMAND_BUFFER_SIZE,
			true, DEFAULT_WAIT_TIME, true, true);

	GetCmdPrompt(wifly);
	//DebugPrint("GotPrompt:");
	//DebugPrint(szWiFlyPrompt);

	WiFlygetDeviceStatus(wifly);

	// try, then try again after reboot.
	if (strlen(wifly->szWiFlyPrompt) < 1) {
		// got a problem
		//DebugPrint(GetBuffer_P(STI_WIFLYDEVICE_ERR_REBOOT, szCmd, SMALL_COMMAND_BUFFER_SIZE));
		reboot(wifly);

		WiFlyDelay(wifly, WIFLY_RESTART_WAIT_TIME);
		// try again

		GetCmdPrompt(wifly);
		if (strlen(wifly->szWiFlyPrompt) < 1) {
			//DebugPrint(GetBuffer_P(STI_WIFLYDEVICE_ERR_START_FAIL, szCmd, SMALL_COMMAND_BUFFER_SIZE));
			bStart = false;
		}
	}

	return bStart;
}

// WiFlygetDeviceStatus
// Refreshes device status flags
//
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
//
// Parameters
// None.
// Returns true on success, false on failure.
//
long WiFlygetDeviceStatus(struct _wifly *wifly) {

	char bufStatus[INDICATOR_BUFFER_SIZE];
	memset(bufStatus, '\0', INDICATOR_BUFFER_SIZE);
	// place leading fixed '8' into first position
	bufStatus[0] = '8';

	ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_STATUS, &bufStatus[1],
			INDICATOR_BUFFER_SIZE - 1, STI_WIFLYDEVICE_GET_STATUS_IND,
			STI_WIFLYDEVICE_RETURN);

	wifly->fStatus = strtol(bufStatus, (char**) 0, 16);
	return wifly->fStatus;
}

// ExtractDetailIdx
// Returns substring from a response.  Indexes point to command, search string and terminator strings.
//
// Parameters:
// idxCommand     StringID of command
// pDetail        pointer to destination buffer
// buflen         length of destination buffer
// idxSearch      StringID to extract AFTER
// idxStop        StringID to extract UNTIL
//
// Returns:       pointer to destination buffer
char* ExtractDetailIdx(struct _wifly *wifly, const int idxCommand,
		char* pDetail, int buflen, const int idxSearch, const int idxStop) {
	char bufCmd[COMMAND_BUFFER_SIZE];
	char bufSearch[INDICATOR_BUFFER_SIZE];
	char bufStop[INDICATOR_BUFFER_SIZE];

	strncpy(bufCmd, WiFlyDevice_string_table[idxCommand], COMMAND_BUFFER_SIZE);
	strncpy(bufSearch, WiFlyDevice_string_table[idxSearch], INDICATOR_BUFFER_SIZE);
	strncpy(bufStop, WiFlyDevice_string_table[idxStop], INDICATOR_BUFFER_SIZE);

	return ExtractDetail(wifly, bufCmd, pDetail, buflen, bufSearch, bufStop);

}

// ExtractDetail
// Returns substring from a response.
//
// Parameters:
// pCommand       pointer to command
// pDetail        pointer to destination buffer
// buflen         length of destination buffer
// pFrom          String to extract AFTER
// pTo            String to extract UNTIL
//
// Returns:       pointer to destination buffer
char* ExtractDetail(struct _wifly *wifly, char* pCommand, char* pDetail,
		int buflen, const char* pFrom, const char* pTo) {
	char* pEndToken = NULL;

	pDetail[0]='\0';
	//send command and ignore results up to after pFrom
	SendCommand(wifly, pCommand, (char*) pFrom, pDetail, buflen, false,
			DEFAULT_WAIT_TIME, false, true);
	// then collect results up to after pTo.
	ScanForPattern(wifly, pDetail, buflen, pTo, true, DEFAULT_WAIT_TIME, true);

	// trim result to not include end token.

	if ((pEndToken = strstr(pDetail, pTo)) != NULL) {
		*pEndToken = '\0';
	}

	// clear buffer of remaining characters
	ScanForPattern(wifly, strchr(pDetail, '\0') + 1,
			buflen - strlen(pDetail) - 1, "\0\0", false, DEFAULT_WAIT_TIME,
			true);
	//   while ( uart.available() > 0 ) {
	//     ch = uart.read();
	//     DebugPrint(ch);
	//   }
	//

	return pDetail;
}

// reboot
// Forces reboot of WiFly
//
// Parameters: none
// Returns: none
#define SOFTWARE_REBOOT_RETRY_ATTEMPTS 5

void reboot(struct _wifly *wifly) {
	char szCommand[SMALL_COMMAND_BUFFER_SIZE];

	strncpy(szCommand, WiFlyDevice_string_table[STI_WIFLYDEVICE_REBOOT],
			SMALL_COMMAND_BUFFER_SIZE);
	// DebugPrint(szCommand);

	if (!WiFlysendCommandSimple(wifly, szCommand, WiFlyFixedPrompts[WIFLY_MSG_AOK])) {
		//DebugPrint( GetBuffer_P(STI_WIFLYDEVICE_ERR_START_FAIL, szCommand, SMALL_COMMAND_BUFFER_SIZE));
		while (1) {
		}; // Hang. TODO: Handle differently?
	}
}

// SendInquiry
// Inquiries provide a device setting result, terminated with a command prompt.
// No specific 'ok/fail' result shown, only ERR or requested response.
// Results placed into global responsebuffer
//
// Parameters:
// Command        The inquiry-command to send
// pBuffer        pointer to a buffer for the response
// bufsize        size of the buffer
//
// Returns true on command success, false on failure.
boolean SendInquiry(struct _wifly *wifly, char *Command, char* pBuffer,
		const int bufsize) {

	return SendCommand(wifly, Command, wifly->szWiFlyPrompt, pBuffer, bufsize,
			true, DEFAULT_WAIT_TIME, true, true);

}
// SendInquirySimple
// Inquiries provide a device setting result, terminated with a command prompt.
// No specific 'ok/fail' result shown, only ERR or requested response.
// Results placed into global responsebuffer
//
// Parameters:
// Command        The inquiry-command to send
//
// Returns true on command success, false on failure.
boolean SendInquirySimple(struct _wifly *wifly, char *Command) {
	char InquiryBuffer[RESPONSE_BUFFER_SIZE];

	boolean bSendInquiry = false;
	bSendInquiry = SendCommand(wifly, Command, wifly->szWiFlyPrompt,
			InquiryBuffer, RESPONSE_BUFFER_SIZE, true, DEFAULT_WAIT_TIME, true,
			true);
	// should trim to returned result less ExpectedPrompt

	return bSendInquiry;

}

// showNetworkScan
// Displays list of available WiFi networks.
//
// Parameters:
// pNetScan    Buffer for scan results (should be large)
// buflen      length of buffer
char* WiflyshowNetworkScan(struct _wifly *wifly, char* pNetScan,
		const int buflen) {

	memcpy(pNetScan, (char) 0, buflen);
	SendCommand(wifly, "scan", "'", pNetScan, buflen, true, SCAN_WAIT_TIME,
			true, false);

	return pNetScan;

}

// setProtocol
// Sets WiFly's communication protocol
// TCP or UDP
//
// Parameters:
// iProtocol      Hex value for protocol - bit mapped values.
// Returns true for success, false for failure or error in call
//
// Note:
// Switching from one to the other requires a reboot of the WiFly.
// For UDP, set all other settings first (unlike example in manual) ; UDP traffic will start upon WiFly reboot.
// This version does not attempt to determine current mode. Reboot forced.
boolean WiFlysetProtocol(struct _wifly *wifly, unsigned int iProtocol) {

	boolean bOk = false;
	unsigned int iMode = 0;
	unsigned int iTmp;
	char bufMode[10];

	iTmp = iProtocol & WIFLY_IPMODE_TCP;

	if (iTmp != 0x00) {
		iMode |= 0x02;
	}

	iTmp = iProtocol & WIFLY_IPMODE_TCP;

	if (iTmp) {
		iMode |= 0x01;
	}

	itoa(iMode, bufMode, 10);

	bOk = issueSetting(wifly, STI_WIFLYDEVICE_SET_PROTOCOL, bufMode);

	// Save settings

	char szCommand[SMALL_COMMAND_BUFFER_SIZE];

	GetBuffer_P(STI_WIFLYDEVICE_SAVE, szCommand, SMALL_COMMAND_BUFFER_SIZE);

	bOk = WiFlysendCommandSimple(wifly, szCommand, WiFlyFixedPrompts[WIFLY_MSG_AOK]);
	reboot(wifly);

	// Allow WiFly to restart
	WiFlyDelay(wifly, WIFLY_RESTART_WAIT_TIME);

	return bOk;
}

// openConnection
// Opens a TCP connection to the provided URL and port (defaults to 80)
//
// Parameters:
// pURL      IP or dns name of server to connect to.
// iWaitTime Time to wait for connection
//
// Returns: true on success, false on failure.
// Remote Port number is set through SetRemotePort
// Note that opened ports can be closed externally / lost connection at any time.
// Opening a connection switches to Data mode from Command mode.
//
// Note: Open and Scan each generate a version-prompt BEFORE results, not after.
boolean open(struct _wifly *wifly, char* bufCommand, const unsigned long iWaitTime ) {
	char bufOpen[INDICATOR_BUFFER_SIZE];
/*
	wifly->bWiFlyConnectionOpen = SendCommand(wifly, bufCommand,
			WiFlyFixedPrompts[WIFLY_MSG_OPEN], bufOpen, INDICATOR_BUFFER_SIZE,
			false, iWaitTime, true, false);
*/
	wifly->bWiFlyConnectionOpen = SendCommand(wifly, bufCommand,
			"", bufOpen, INDICATOR_BUFFER_SIZE,
			true, iWaitTime, true, true);

	if (wifly->bWiFlyConnectionOpen) {
		wifly->bWiFlyInCommandMode = false;
	}

	return wifly->bWiFlyConnectionOpen;
}

boolean WiFlyopenConnection(struct _wifly *wifly, const char* pURL,
		const unsigned long iWaitTime) {
	char bufCommand[COMMAND_BUFFER_SIZE];

	memset(bufCommand, '\0', COMMAND_BUFFER_SIZE);
	GetBuffer_P(STI_WIFLYDEVICE_OPEN, bufCommand, COMMAND_BUFFER_SIZE);
	strcat(bufCommand, pURL);
	strcat(bufCommand, " ");
	itoa(wifly->iRemotePort, strchr(bufCommand, '\0'), 10);
	return open(wifly, bufCommand, iWaitTime);
}


// closeConnection
// closes an open connection
//
// Parameters:
// bSafeClose	   (default) Use slow 'safe close'
//                 false if high confidence connection is in fact open,
//				   and can tolerate occasional errors if connection
//                 closes sooner than expected.
//
// returns   true on close, false on failure.
// Side effects: bWiFlyConnectionOpen should become false
//          and  bWiFlyInCommandMode should become true.
//
// Notes:
// 1. When closed via a command the WiFly state is in command mode.
// 2. External connection close leaves WiFly in data mode
// 3. Closing a closed connection results in an error.
// 4. External connection close could happen at any time.
// 5. WiFlygetDeviceStatus() gives actual WiFly status at the moment of calling.
// 6. When opened via listening, info from previous WiFlygetDeviceStatus() is out-of-date.
// 7. The 'close' command results with a command-prompt, no 'AOK' or similar.
//     then, a *CLOS* signal appears.
//
// returns true on closure, false on failure to close.
//
boolean WiFlycloseConnection(struct _wifly *wifly, boolean bSafeClose) {
	// if a connection is open then close it.
	if (wifly->bWiFlyConnectionOpen) {
		// first see if connection is *STILL* open.
		boolean bTrySafeClose = bSafeClose;
		boolean bDoClose = true;

		// repeat until closed...
		while (wifly->bWiFlyConnectionOpen) {
			if (bTrySafeClose) {
				WiFlygetDeviceStatus(wifly);
				if (WiFlyisTCPConnected(wifly) && wifly->bWiFlyConnectionOpen) {
					bDoClose = true;
				} else {
					bDoClose = false;
					wifly->bWiFlyConnectionOpen = false; // should be redundant as a caught *CLOS* will set this to false.
				}
			} // if doing safe close
			if (bDoClose) {
				WiFlydrain(wifly);
				char bufCmd[INDICATOR_BUFFER_SIZE];
				char bufClose[INDICATOR_BUFFER_SIZE];
				memset(bufCmd, '\0', INDICATOR_BUFFER_SIZE);
				memset(bufClose, '\0', INDICATOR_BUFFER_SIZE);

				// close command response is a prompt, then a *CLOS* signal after.
				SendCommand(wifly,
						    GetBuffer_P(STI_WIFLYDEVICE_CLOSE, bufCmd, INDICATOR_BUFFER_SIZE),
							WiFlyFixedPrompts[WIFLY_MSG_PROMPT], bufClose,
							INDICATOR_BUFFER_SIZE, true, DEFAULT_WAIT_TIME, false, true);
				WiFlydrain(wifly);
				wifly->bWiFlyInCommandMode = true;
			}
		} // while
	} else {
		// closed externally already.
		// No change to bWiFlyCommandMode.
		return true;
	}
	return true;
}

// drain
// Empties incoming buffer

int WiFlydrain(struct _wifly *wifly) {
	char chDrain;
	unsigned long TimeOutTime = wifly->millis() + DEFAULT_WAIT_TIME;
	// DebugPrint("Waiting for signal");
	while (wifly->bWiFlyConnectionOpen && UARTisAvailable(&(wifly->uart)) > 0 && wifly->millis() < TimeOutTime) {
		UARTreadByte(&(wifly->uart), (byte *) &chDrain);
	}
	// DebugPrint("Drained.");
	return 1;
}

// serveConnection
// Waits for a client to connect on the given port.
//
// Parameters:
// reconnectWaitTime  Duration to wait before verifying wlan and reconnecting if needed.
//
// returns            true on connection, false on internal failure.
//
boolean WiFlyserveConnection(struct _wifly *wifly,
		const unsigned long reconnectWaitTime) {
	char bufRequest[COMMAND_BUFFER_SIZE];
	int iRequest;
	boolean bReturn = false;
	int mask = PROMPT_EXPECTED_TOKEN_FOUND | PROMPT_OPEN | PROMPT_OPEN_ALREADY;

	iRequest = ScanForPattern(wifly, bufRequest, COMMAND_BUFFER_SIZE,
			WiFlyFixedPrompts[WIFLY_MSG_OPEN], false, reconnectWaitTime, true);
	if ((iRequest & mask) != 0 ) {
		//memset (bufRequest,'\0',COMMAND_BUFFER_SIZE);
		wifly->bWiFlyInCommandMode = false;
		bReturn = true;
	} else {
		// No connection within timeout.
		bReturn = false;
		// reconnection logic
	}

	// TODO: decision on timeout/reconnection status


	return bReturn;
}

// getMAC
// Returns WiFly MAC address
//
// Parameters:
// bufMAC    buffer for MAC address
// buflen    length of buffer (should be at least 18 chars)
// Returns:  pointer to supplied buffer MAC address, or empty string on failure.
// Format expected: Mac Addr=xx:xx:xx:xx:xx:xx
char* WiFlygetMAC(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_MAC, pbuf, buflen,
			STI_WIFLYDEVICE_GET_MAC_ADDR, STI_WIFLYDEVICE_RETURN);
}

// getIP
// Returns WiFly IP address
//
// Parameters
// bufIP     buffer for IP address
// buflen    length of buffer (should be at least 18 chars for IPv4 and longer for IPv6)
// Returns:  pointer to supplied buffer with IP address, will be empty string on failure.
//
char* WiFlygetIP(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_IP_DETAILS, pbuf, buflen,
			STI_WIFLYDEVICE_GET_IP_IND, STI_WIFLYDEVICE_RETURN);
}

// getNetMask
// Returns WiFly Netmask
//
// Parameters
// buf     buffer for Netmask
// buflen    length of buffer (should be at least 18 chars for IPv4 and longer for IPv6)
// Returns:  pointer to supplied buffer with GW address, will be empty string on failure.
char* WiFlygetNetMask(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_IP_DETAILS, pbuf, buflen,
			STI_WIFLYDEVICE_GET_NM_IND, STI_WIFLYDEVICE_RETURN);
}

// getGateway
// Returns WiFly Gateway address
//
// Parameters
// bufGW     buffer for IP address
// buflen    length of buffer (should be at least 18 chars for IPv4 and longer for IPv6)
// Returns:  pointer to supplied buffer with GW address, will be empty string on failure.
char* WiFlygetGateway(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_IP_DETAILS, pbuf, buflen,
			STI_WIFLYDEVICE_GET_GW_IND, STI_WIFLYDEVICE_RETURN);
}

// getDNS
// Returns WiFly DNS address
//
// Parameters
// bufDNS    buffer for IP address
// buflen    length of buffer (should be at least 18 chars for IPv4 and longer for IPv6)
// Returns:  pointer to supplied buffer with DNS address, will be empty string on failure.
char* WiFlygetDNS(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_DNS_DETAILS, pbuf, buflen,
			STI_WIFLYDEVICE_GET_DNS_IND, STI_WIFLYDEVICE_RETURN);

}

// getSSID
// Returns current SSID
//
// Parameters
// bufSSID   buffer for SSID
// buflen    length of buffer (should be at least 18 chars for IPv4 and longer for IPv6)
// Returns:  pointer to supplied buffer with SSID, will be empty string on failure.
//
char* WiFlygetSSID(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_WLAN, pbuf, buflen,
			STI_WIFLYDEVICE_GET_WLAN_SSID_IND, STI_WIFLYDEVICE_RETURN);
}

// getRSSI
// Returns current RSSI
//
// Parameters
// pbuf      buffer for RSSI
// buflen    length of buffer
// Returns:  pointer to supplied buffer with RSSI, will be empty string on failure.
//
char* WiFlygetRSSI(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_RSSI, pbuf, buflen,
			STI_WIFLYDEVICE_GET_RSSI_IND, STI_WIFLYDEVICE_RETURN);
}

// getBattery
// Returns current Battery voltage
//
// Parameters
// pbuf      buffer for voltage
// buflen    length of buffer
// Returns:  pointer to supplied buffer with battery voltage, will be empty string on failure.
//
char* WiFlygetBattery(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_BATTERY, pbuf, buflen,
			STI_WIFLYDEVICE_GET_BATTERY_IND, STI_WIFLYDEVICE_RETURN);
}

// getDeviceID
// Returns current DeviceID
//
// Parameters
// bufSSID   buffer for DeviceID
// buflen    length of buffer (should be at least 18 chars for IPv4 and longer for IPv6)
// Returns:  pointer to supplied buffer with DeviceID, will be empty string on failure.
//
char* WiFlygetDeviceID(struct _wifly *wifly, char* pbuf, int buflen) {
	return ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_WLAN, pbuf, buflen,
			STI_WIFLYDEVICE_GET_WLAN_DEV_IND, STI_WIFLYDEVICE_RETURN);
}

// getTxPower
// Returns transmit power.
//
// Parameters:
// Returns:  integer with the actual transmission power
// Format expected: TxPower=XX
int WiFlygetTxPower(struct _wifly *wifly) {
	char bufTxPower[COMMAND_BUFFER_SIZE];

	return (atoi(
			ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_WLAN, bufTxPower,
					COMMAND_BUFFER_SIZE, STI_WIFLYDEVICE_GET_WLAN_TX_IND,
					STI_WIFLYDEVICE_RETURN)));
}

// getTime
// Returns (in milliseconds) time since bootup or Unix epoch if NTP server updated.
//
// Parameters:
// Returns:  unsigned long representing seconds since Unix Epoch or power-on.
// Format expected: RTC=tttttttt ms
unsigned long WiFlygetTime(struct _wifly *wifly) {
	char bufTimeSerial[COMMAND_BUFFER_SIZE];

	return (atol(
			ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_TIME, bufTimeSerial,
					COMMAND_BUFFER_SIZE, STI_WIFLYDEVICE_GET_TIME_IND,
					STI_WIFLYDEVICE_RETURN)) + wifly->lUTC_Offset_seconds);
}

// TODO: add GetTimeStatus


// isTCPConnected
//
// Returns true if TCP Connected, false if not.
// Dependent on prior call to WiFlygetDeviceStatus();
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
// bit 1
//
// Parameters
// None.
// Returns true on success, false on failure.
//
boolean WiFlyisTCPConnected(struct _wifly *wifly) {

	if ((wifly->fStatus & 0x8000) == 0) {
		WiFlygetDeviceStatus(wifly);
	}

	return ((wifly->fStatus & 0x01) > 0 ? true : false);

}

// isAssociated
//
// Returns true if associated with a wifi SSID, false if not.
// Dependent on prior call to WiFlygetDeviceStatus();
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
// bit 4
//
// Parameters
// None.
// Returns true on success, false on failure.
//
boolean WiFlyisAssociated(struct _wifly *wifly) {

	if ((wifly->fStatus & 0x8000) == 0) {
		WiFlygetDeviceStatus(wifly);
	}

	return ((wifly->fStatus & 0x10) > 0 ? true : false);

}

// isAuthenticated
//
// Returns true if authenticated with a wifi SSID, false if not.
// Dependent on prior call to WiFlygetDeviceStatus();
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
// bit 5
//
// Parameters
// None.
//
boolean WiFlyisAuthenticated(struct _wifly *wifly) {

	if ((wifly->fStatus & 0x8000) == 0) {
		WiFlygetDeviceStatus(wifly);
	}

	return ((wifly->fStatus & 0x20) > 0 ? true : false);

}

// isDNSfound
//
// Returns true if DNS is found, false if not.
// Dependent on prior call to WiFlygetDeviceStatus();
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
// bit 6
//
// Parameters
// None.
//
boolean WiFlyisDNSfound(struct _wifly *wifly) {

	if ((wifly->fStatus & 0x8000) == 0) {
		WiFlygetDeviceStatus(wifly);
	}

	return ((wifly->fStatus & 0x40) > 0 ? true : false);

}

// isDNScontacted
//
// Returns true if DNS is contacted, false if not.
// Dependent on prior call to WiFlygetDeviceStatus();
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
// bit 7
//
// Parameters
// None.
//
boolean WiFlyisDNScontacted(struct _wifly *wifly) {
	if ((wifly->fStatus & 0x8000) == 0) {
		WiFlygetDeviceStatus(wifly);
	}

	return ((wifly->fStatus & 0x80) > 0 ? true : false);

}

// getChannel
//
// Returns wifi channel number .
// Dependent on prior call to WiFlygetDeviceStatus();
// WiFly 'show connection' command gives a result in hex 8XYZ as a string.
// bits 9-
//
// Parameters
// None.
// Returns true on success, false on failure.
//
int WiFlygetChannel(struct _wifly *wifly) {

	if ((wifly->fStatus & 0x8000) == 0) {
		WiFlygetDeviceStatus(wifly);
	}

	return ((wifly->fStatus >> 8) - 0x80);

}

// isInCommandMode
//
// Returns true if WiFly in Command Mode false if not.
//
// Parameters
// None.
//
boolean WiFlyisInCommandMode(struct _wifly *wifly) {

	return wifly->bWiFlyInCommandMode;

}

// isConnectionOpen
//
// Returns true if an IP connection is open, false if not.
//
// Parameters
// None.
//
boolean WiFlyisConnectionOpen(struct _wifly *wifly) {

	return wifly->bWiFlyConnectionOpen;

}

// getLibraryVersion
// Returns WiFly Driver Library version
//
// Parameters
// pDetail        pointer to destination buffer
// buflen         length of destination buffer
// Returns:  pointer to supplied buffer with library version; empty string on failure.
//
char* WiFlygetLibraryVersion(struct _wifly *wifly, char* pbuf, int buflen) {
	return GetBuffer_P(STI_WIFLYDEVICE_LIBRARY_VERSION, pbuf, buflen);
}

// isifUp
// Returns true if currently connected to Access Point.
//
// Parameters: None
// Returns:  IP netmask, or empty string on failure.
boolean WiFlyisifUp(struct _wifly *wifly) {
	boolean bReturn = false;
	char buf[COMMAND_BUFFER_SIZE];

	ExtractDetailIdx(wifly, STI_WIFLYDEVICE_GET_IP_DETAILS, buf, COMMAND_BUFFER_SIZE,
			STI_WIFLYDEVICE_GET_IP_UP_IND, STI_WIFLYDEVICE_RETURN);
	bReturn = (strcmp(buf, "UP") == 0 ? true : false);

	return bReturn;

}

// ExtractLineFromBuffer
// Returns string extracted from provided buffer.
//
//
// Parameters:
// idString         StringID of command to send
// pBuffer          pointer to provided buffer
// bufsize          expected size of required buffer
// pStartPattern    pointer to null-terminated string identifying the start of desired line
// chTerminator     character to terminate desired line.

// Returns:  pointer to within provided buffer, with result or empty string on failure.
char* ExtractLineFromBuffer(struct _wifly *wifly, const int idString,
		char* pBuffer, const int bufsize, const char* pStartPattern,
		const char* pTerminator) {
	char* pStart;
	char* pTerm;
	boolean bOk = false;
	char szCommand[SMALL_COMMAND_BUFFER_SIZE];
	char* pResponse = pBuffer;

	if (!SendInquiry(wifly,
			GetBuffer_P(idString, szCommand, SMALL_COMMAND_BUFFER_SIZE),
			pBuffer, bufsize)) {
		bOk = false;
	} else {
		pStart = strstr(pBuffer, pStartPattern);
		if (pStart != NULL) {
			// Move pointer past the start pattern
			pStart += strlen(pStartPattern);
			pResponse = pStart;
		}
		pTerm = strstr(pStart, pTerminator);
		if (pTerm == NULL) {
			bOk = false;
		} else {
			*(pTerm) = '\0';
			bOk = true;
		} // if end-of-line found

	}
	if (!bOk) {
		*(pResponse) = '\0';
	}
	return pResponse;
}

// leave
// Disconnects from current WIFI lan
//
// Parameters:
//
// Returns true on command success, false on failure.
boolean WiFlyleave(struct _wifly *wifly) {

	boolean bSendLeave = false;
	char szCmd[COMMAND_BUFFER_SIZE];
	char szReply[INDICATOR_BUFFER_SIZE];

	bSendLeave
	= SendCommand(wifly,
			GetBuffer_P(STI_WIFLYDEVICE_LEAVE, szCmd,
					COMMAND_BUFFER_SIZE),
					GetBuffer_P(STI_WIFLYDEVICE_DEAUTH, szReply,
							INDICATOR_BUFFER_SIZE), szCmd, COMMAND_BUFFER_SIZE,
							false, DEFAULT_WAIT_TIME, true, true);

	return bSendLeave;

}

boolean WiFlysetSSID(struct _wifly *wifly, const char* pSSID) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_SSID, pSSID);

}

boolean WiFlysetPassphrase(struct _wifly *wifly, const char* pPassphrase) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_PASSPHRASE, pPassphrase);

}

// Sets NTP server address
boolean WiFlysetNTP(struct _wifly *wifly, const char* pNTP) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_NTP, pNTP);

}

// Sets frequency for NTP updates
boolean WiFlysetNTP_Update_Frequency(struct _wifly *wifly,
		const char* pNTP_Update) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_NTP_ENABLE, pNTP_Update);
}

// sets offset from UTC.
// Parameters:
// fltUTC_Offset_Hours  Number of hours from UTC  (e.g. PST = -8 )
// Setting is used during setTime().
// returns true always.
boolean WiFlysetNTP_UTC_Offset(struct _wifly *wifly, float fltUTC_Offset_hours) {
	wifly->lUTC_Offset_seconds = (long) (fltUTC_Offset_hours * 60 * 60);
	return true;
}
// Sets WiFly DNS name
boolean WiFlysetDeviceID(struct _wifly *wifly,const char* pDeviceID) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_DEVICEID, pDeviceID);

}

// issueSetting
// Issues a WiFly setting command
// Parameters:
// IdxCommand    Index into string table of command
// pParam        null-terminated string of parameter.
//               Command and parameter must be less than COMMAND_BUFFER_SIZE
//
// Returns - true on Command success, false on fail.
boolean issueSetting(struct _wifly *wifly, int idxCommand, const char* pParam) {

	char szReply[INDICATOR_BUFFER_SIZE];
	//  char szIndicator[INDICATOR_BUFFER_SIZE];
	char szBuffer[COMMAND_BUFFER_SIZE];

	GetBuffer_P(idxCommand, szBuffer, COMMAND_BUFFER_SIZE);
	strncat(szBuffer, pParam, COMMAND_BUFFER_SIZE - strlen(szBuffer));

	return SendCommand(wifly, szBuffer, WiFlyFixedPrompts[WIFLY_MSG_AOK], szReply,
			INDICATOR_BUFFER_SIZE, true, DEFAULT_WAIT_TIME, true, true);
}

// SetUseDCHP
// Sets DHCP to requested mode
boolean WiFlysetDHCPMode(struct _wifly *wifly, const int iDHCPMode) {
	char bufMode[10];

	itoa(iDHCPMode, bufMode, 10);

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_DHCP, bufMode);
}

// SetIP
// Sets static IP address
// Parameters:
// pIP      null-terminated character string of the IP address e.g. '192.168.1.3'
boolean WiFlysetIP(struct _wifly *wifly, const char* pIP) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_IP, pIP);
}

// SetNetMask
// Sets static IP netmask
// Parameters:
// pNM      null-terminated character string of the netmask e.g. '255.255.255.0'
boolean WiFlysetNetMask(struct _wifly *wifly, const char* pNM) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_NETMASK, pNM);
}

// SetGateway
// Sets static Gateway address
// Parameters:
// pGW      null-terminated character string of the Gateway address e.g. '192.168.1.254'
boolean WiFlysetGateway(struct _wifly *wifly, const char* pGW) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_GATEWAY, pGW);
}

// SetDNS
// Sets static DNS address
// Parameters:
// pDNS      null-terminated character string of the DNS address e.g. '192.168.1.1'
boolean WiFlysetDNS(struct _wifly *wifly, const char* pDNS) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_DNS, pDNS);
}

// SetChannel
// Sets wifi Channel
// Parameters:
// pChannel      null-terminated character string of the channel e.g. '6'
boolean WiFlysetChannel(struct _wifly *wifly, const char* pChannel) {

	return issueSetting(wifly, STI_WIFLYDEVICE_SET_CHANNEL, pChannel);
}

// SetAuthMode
// Sets wifi Authentication mode
// Parameters:
// iAuthMode      Authentication mode of type WIFLY_AUTH_XXXX
boolean WiFlysetAuthMode(struct _wifly *wifly, int iAuthMode) {
	char bufMode[10];

	itoa(iAuthMode, bufMode, 10);
	return issueSetting(wifly, STI_WIFLYDEVICE_SET_WIFI_AUTH, bufMode);
}

// SetJoinMode
// Sets wifi Network Join mode
// Parameters:
// iJoinMode     join mode of type WIFLY_JOIN_XXXXXX

boolean WiFlysetJoinMode( struct _wifly *wifly, int iJoinMode) {
	char bufMode[10];

	itoa( iJoinMode, bufMode, 10);
	return issueSetting( wifly, STI_WIFLYDEVICE_SET_WIFI_JOIN, bufMode );
}

// SetLocalPort
// Sets Local port for listening
// Parameters:
// iLocalPort    Port to listen to.
boolean WiFlysetLocalPort(struct _wifly *wifly, int iNewLocalPort) {
	char bufPort[10];

	itoa(iNewLocalPort, bufPort, 10);
	wifly->iLocalPort = iNewLocalPort;
	return issueSetting(wifly, STI_WIFLYDEVICE_SET_LOCAL_PORT, bufPort);
}

// SetRemotePort
// Sets Remote port for connections
// Parameters:
// iRemotePort    Port to contact on remote server.
boolean WiFlysetRemotePort(struct _wifly *wifly, int iNewRemotePort) {
	char bufPort[10];

	itoa(iNewRemotePort, bufPort, 10);
	wifly->iRemotePort = iNewRemotePort;
	return issueSetting(wifly, STI_WIFLYDEVICE_SET_REMOTE_PORT, bufPort);
}

// SetTxPower
// Sets transmit power
// Parameters:
// iNewPower    Transmission power to set.
boolean WiFlysetTxPower( struct _wifly  *wifly, int iNewPower) {
	char bufPower[10];

	// Value must be between 0 and 12
	if (iNewPower < 0) iNewPower=0;
	if (iNewPower > 12) iNewPower=12;

	itoa(iNewPower, bufPower, 10);
	return issueSetting(wifly, STI_WIFLYDEVICE_SET_WLAN_TX, bufPower);
}

// join
// Parameters: None
// Joins a network with previously supplied setSSID and passphrase.
//
// returns true on success, false on failure
boolean WiFlyjoin(struct _wifly *wifly) {

	char szSSID[COMMAND_BUFFER_SIZE];
	WiFlygetSSID(wifly, szSSID, COMMAND_BUFFER_SIZE);
	return WiFlyjoinSSID(wifly, szSSID);
}

// join
// Parameters: None
// Joins a network with given SSID and previously-provided passphrase.
//
// returns true on success, false on failure.
// Todo: support spaces in passphrase.
boolean WiFlyjoinSSID(struct _wifly *wifly, char* pSSID) {

	boolean bJoined = false;
	char szCmd[COMMAND_BUFFER_SIZE];
	WiFlysetSSID(wifly, pSSID);
	GetBuffer_P(STI_WIFLYDEVICE_INDEX_JOIN, szCmd, COMMAND_BUFFER_SIZE);
	strncat(szCmd, pSSID, COMMAND_BUFFER_SIZE - strlen(szCmd));

	char bufIndicator[INDICATOR_BUFFER_SIZE];
	char bufResponse[RESPONSE_BUFFER_SIZE];

	bJoined = SendCommand(wifly,
			szCmd,
			GetBuffer_P(STI_WIFLYDEVICE_INDEX_ASSOCIATED, bufIndicator,
					INDICATOR_BUFFER_SIZE), bufResponse, RESPONSE_BUFFER_SIZE,
					true, JOIN_WAIT_TIME, false, false);

	return bJoined;

}

char * WiFlysendPing(struct _wifly *wifly, char* host, char* pbufResponse, int buflen) {
	char szCmd[strlen(host)+10];
	strcpy(szCmd, "ping ");
	strcat(szCmd, host);
	pbufResponse[0]='\0';
	SendCommand(wifly,
			szCmd,
				"Ping try", pbufResponse, buflen,
						true, JOIN_WAIT_TIME, false, false);
	return pbufResponse;
}

/*
 * Sends an HTTP GET to defined host/ip
 *
 *
 */
unsigned long WiFlygetDataHTTPGet(struct _wifly *wifly, char* urlOrg, char* pbufResponse, int buflen, uint32_t waitTime) {
	char szCmd[COMMAND_BUFFER_SIZE];
	char *root;
	char *url = urlOrg;
	char hostBuf[64];
	char *host = hostBuf;
	unsigned long bytesReaded = 0L;
	unsigned long contentLength = 0L;

	int iResponse = 0;
	char *close = NULL;
	boolean success = false;
	boolean binary = false;
	int pass = 0;


	if( !strncmpi(url, "http://", 7) ) url+=7; // remove "http://" if exist

	root = strstr(url, "/");
	if( root == NULL ) {
		root="/";
		strcpy(host, url);
	} else {
		close = url;
		while(close != root){
			*host=*close;
			host++;
			close++;
		}
		*host='\0';
		host = hostBuf;
	}


	// Build open command from URL ("open " + host or ip + " 80")

	pbufResponse[0]='\0';
	strcpy(pbufResponse, "set dns name ");
	strcat(pbufResponse, host); // host

	SendCommand(wifly, pbufResponse,
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);

	// Disable IP
	SendCommand(wifly, "set ip host 0",
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);

	// Remote port
	SendCommand(wifly, "set ip remote 80",
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);

	// TCP only
	SendCommand(wifly, "set ip proto 2", //18
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);
	// Don't send headers (we'll do it)
	SendCommand(wifly, "set option format 0", //1
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);
	// Dont set remote message
	SendCommand(wifly, "set com remote 0",
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);

/*
	// Send data on connect GET + root
	SendCommand2(wifly, "set com remote GET$", root,
			WiFlyFixedPrompts[WIFLY_MSG_AOK], szCmd, COMMAND_BUFFER_SIZE,
			true, JOIN_WAIT_TIME, true, true);
*/

	UARTflush(&(wifly->uart));


	// Open connection
	//	SendCommand(wifly, pbufResponse,
	//			WiFlyFixedPrompts[WIFLY_MSG_OPEN], szCmd, COMMAND_BUFFER_SIZE,
	//			true, JOIN_WAIT_TIME, true, true);

	if (StartCommandMode(wifly, pbufResponse, buflen)) {
		UARTsendString(&(wifly->uart), (byte*) "open");
		UARTsendString(&(wifly->uart), (byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_RETURN]);
		UARTflush(&(wifly->uart));


		//... and receive results.
		WiFlyDelayUntilNoData(wifly, waitTime);
		WiFlyDelayUntilReceivingData(wifly, 50);

		iResponse = ScanForSinglePattern(wifly, pbufResponse, buflen,
				WiFlyFixedPrompts[WIFLY_MSG_OPEN], true, waitTime, true);

		UARTsendString(&(wifly->uart), (byte*) "GET ");
		UARTsendString(&(wifly->uart), (byte*) root);
		UARTsendString(&(wifly->uart), (byte*) " HTTP/1.1\nHost: ");
		UARTsendString(&(wifly->uart), (byte*) host);
		UARTsendString(&(wifly->uart), (byte*) "\nAccept: */*\nConnection: close\n");
		UARTsendString(&(wifly->uart), (byte*) "\n");

		DebugPrintSend(wifly, root, DEBUG_MODE_INFO);
		WiFlyDelayUntilNoData(wifly, waitTime);


		if(UARTisAvailable(&(wifly->uart))) {
			do {
				close = pbufResponse;
				*close = '\0';
				// Read a line

				do {
					if(UARTisAvailable(&(wifly->uart))) {
						UARTreadByte(&(wifly->uart), (byte *) close);
						if(*close == '\r') *close = '\0';
						if(*close == '\n') {
							break;
						}
						close++;
						pass=0;
					} else {
						if(pass > 5) break; // Wait for five times of 50 ms.
						WiFlyDelay(wifly, 50);
						pass++;
					}
				}while(1);
				*close = '\0';

				// doesn't work without a delay
				// UARTreadLine(&(wifly->uart), (byte *) pbufResponse, (uint32_t) buflen);
				DebugPrintReceive(wifly, pbufResponse, DEBUG_MODE_INFO);
				if( strstr(pbufResponse, "200 OK") ) success = true;
				else if( strstr(pbufResponse, "Content-Type:") ){
					if( strstr(pbufResponse, "octet-stream") ) binary = true;
				}
				else if( strstr(pbufResponse, "Content-Length:") ) {
					close=strstr(pbufResponse, ":");
					if( close != NULL ){
						close ++;
						contentLength=atol(close);
					}
				}
			} while(strlen(pbufResponse) != 0);


			close = pbufResponse;
			if(success) { // On success, read data on buffer
				do {
					pass=0;
					do {
						if(UARTisAvailable(&(wifly->uart))) {
							UARTreadByte(&(wifly->uart), (byte *) close);
							close++;
							bytesReaded++;
							buflen--;
							contentLength--;
							pass=0;
							break;
						} else {
							if(pass > 5) break; // Wait for five times of 50 ms.
							WiFlyDelay(wifly, 50);
							pass++;
						}
					}while(true);
				} while( (contentLength > 0) && (buflen > 0) && !pass);
				DebugPrintReceive(wifly, pbufResponse, DEBUG_MODE_INFO);
			} else {
				UARTflush(&(wifly->uart));
			}

		} else {// There is not response
		}
		/*
		if( (iResponse & PROMPT_EXPECTED_TOKEN_FOUND)== PROMPT_EXPECTED_TOKEN_FOUND)  {
			pbufResponse[0]='\0';
			iResponse = ScanForSinglePattern(wifly, pbufResponse, buflen,
					WiFlyFixedPrompts[WIFLY_MSG_CLOSE], true, JOIN_WAIT_TIME, true);
		}
		if( (close=strstr(pbufResponse, WiFlyFixedPrompts[WIFLY_MSG_CLOSE]))) {
		 *close='\0';
		}
		 */
	}
	return bytesReaded;
}




/*
 * Sends an HTTP GET to defined host/ip
 *
 *
 */
boolean WiFlysendHTTPGet(struct _wifly *wifly, char* url, char* pbufResponse, int buflen, uint32_t waitTime) {
	char szCmd[200];
	int iResponse = 0;
	char *close = NULL;
	boolean result = false;

//	WiFlysetProtocol(wifly, 18);

	SendCommand(wifly, "set ip proto 18",
			    WiFlyFixedPrompts[WIFLY_MSG_AOK], pbufResponse, buflen,
						true, JOIN_WAIT_TIME, true, true);


	SendCommand(wifly, "set option format 1",
			    WiFlyFixedPrompts[WIFLY_MSG_AOK], pbufResponse, buflen,
						true, JOIN_WAIT_TIME, true, true);

	strcpy(szCmd, "set comm remote GET$");
	strcat(szCmd, url);
	pbufResponse[0]='\0';
	SendCommand(wifly,
			    szCmd,
			    WiFlyFixedPrompts[WIFLY_MSG_AOK], pbufResponse, buflen,
						true, JOIN_WAIT_TIME, true, true);



	pbufResponse[0]='\0';

	UARTflush(&(wifly->uart));

/*
 * Open connection and wait for a response beetwen *OPEN* and *CLOS*
 */

	if (StartCommandMode(wifly, pbufResponse, buflen)) {
		UARTsendString(&(wifly->uart), (byte*) "open");
		UARTsendString(&(wifly->uart), (byte*) WiFlyDevice_string_table[STI_WIFLYDEVICE_RETURN]);
		UARTflush(&(wifly->uart));

		WiFlyDelayUntilNoData(wifly, waitTime);
		WiFlyDelayUntilReceivingData(wifly, 50);

		iResponse = ScanForSinglePattern(wifly, pbufResponse, buflen,
				WiFlyFixedPrompts[WIFLY_MSG_OPEN], true, JOIN_WAIT_TIME, true);

		if( (iResponse & PROMPT_EXPECTED_TOKEN_FOUND)== PROMPT_EXPECTED_TOKEN_FOUND)  {
			pbufResponse[0]='\0';
			iResponse = ScanForSinglePattern(wifly, pbufResponse, buflen,
								WiFlyFixedPrompts[WIFLY_MSG_CLOSE], true, JOIN_WAIT_TIME, true);
		}
		if( (close=strstr(pbufResponse, WiFlyFixedPrompts[WIFLY_MSG_CLOSE]))) {
			*close='\0';
		}
	}

	if(pbufResponse[0]!='\0') {
		result=true;
	}
	return result;
}



char *WiFlygetFixedPrompt(const int PromptIndex)
{
	return (char*) WiFlyFixedPrompts[PromptIndex];
}

char *WiFlygetDeviceString(const int StringIndex)
{
	return (char*) WiFlyDevice_string_table[StringIndex];
}

void DebugPrint( struct _wifly  *wifly, char* pParam, int mode)
{
	if( wifly->bDebugMode >= mode && wifly->Debug != NULL) {
		wifly->Debug(wifly->debugParam, pParam, strlen(pParam), mode);
	}
}

void DebugPrintSend( struct _wifly  *wifly, char* pParam, int mode)
{
	if( wifly->bDebugMode >= mode && wifly->Debug != NULL) {
		wifly->Debug(wifly->debugParam, "-> ", 3, mode);
		wifly->Debug(wifly->debugParam, pParam, strlen(pParam), mode);
		wifly->Debug(wifly->debugParam, "\n", 1, mode);
	}
}

void DebugPrintReceive( struct _wifly  *wifly, char* pParam, int mode)
{
	if( wifly->bDebugMode >= mode && wifly->Debug != NULL) {
		wifly->Debug(wifly->debugParam, "<- ", 3, mode);
		wifly->Debug(wifly->debugParam, pParam, strlen(pParam), mode);
		wifly->Debug(wifly->debugParam, "\n", 1, mode);
	}
}

void DebugPrintChar( struct _wifly  *wifly, char* pParam, int len, int mode)
{
	if( wifly->bDebugMode > mode && wifly->Debug != NULL) {
		wifly->Debug(wifly->debugParam, pParam, len, mode);
	}
}
